// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <iomanip>
#include "Solver.hpp"
#include "boost/lexical_cast.hpp"
#include <algorithm>

#ifndef NOSIGNALS

#include <signal.h>

static bool statsRequest = false;
static bool traceRequest = false;
static bool sigRequest   = false;

void setStatsRequest(int sig)
{
    statsRequest = true;
    sigRequest   = true;
}

void setTraceRequest(int sig)
{
    traceRequest = true;
    sigRequest   = true;
}

#endif

using boost::lexical_cast;

Solver::Solver()
    : 
        fila(0),
        fitFilter(FILTER_OFF),
        goal(0),
        info(false),
        parityBacktrack(false),
        parityFilter(FILTER_OFF),
        quiet(false),
        redundancyFilterName(REDUNDANCY_FILTER_AUTO_NAME),
        redundancyFilterIndex(REDUNDANCY_FILTER_AUTO_INDEX),
        trace(0),
        unique(false),
        volumeBacktrack(0),
        volumeFilter(FILTER_OFF),
        volumeBacktrackControl(0),
        fitFilterControl(0),
        parityFilterControl(0),
        volumeFilterControl(0),
        filterControl(0),
        parityPlaceControl(0),
        volumePlaceControl(0),
        uniqueControl(false),
        endTrialControl(false),
        puzzle(NULL),
        numMobilePieces(0),
        remainingShapeList(NULL),
        remainingShapeCount(NULL),
        traceCount(0),
        solutionCount(0),
        redundantCount(0),
#ifndef NO_POLY_STATS
        parityBacktrackCount(0),
        parityFilterCount(NULL),
        volumeBacktrackCount(NULL),
        volumeFilterCount(NULL),
        fitFilterCount(NULL),
        attemptsList(NULL),
        fitsList(NULL),
#endif
        solveMeter(NULL),
        solveInitMeter(NULL),
        solveAlgoMeter(NULL),
        solveFilterMeter(NULL),
        filaMeter(NULL),
        filaInitMeter(NULL),
        filaAlgoMeter(NULL),
        filaCleanupMeter(NULL),
        solveCleanupMeter(NULL)
{
    solveMeter = PolyPerf::getInstance()->getMeter("solve", "all");
}


void Solver::solve(Puzzle& puzzle)
{
    PerformanceMeasurement measure(solveMeter);

    this->puzzle.setPuzzle(&puzzle);
    init();
    solve();
    cleanup();
}


void Solver::init()
{
    if(solveInitMeter == NULL)
        solveInitMeter = PolyPerf::getInstance()->getMeter("solveInit", "solve");

    PerformanceMeasurement measure(solveInitMeter);

    numMobilePieces = puzzle.getNumMobilePieces();
    allocRemainingShapeList();
    initStats();
    redundancyFilterIndex = redundancyFilterNameToIndex();
    redundancyFilterIndex = puzzle.genImageLists(redundancyFilterIndex);
    initControlVariables();

    // Initialize the parity and control monitors only if we intend to use
    // them.  In particular, volume monitor initialization can allocate
    // enormous amounts of memory if puzzle pieces have many different sizes.
    // Thanks to Andrew Juell for discovering this problem.
    //
    if(parityPlaceControl <= numMobilePieces)
        puzzle.initParityMonitor();
    if(volumePlaceControl <= numMobilePieces)
        puzzle.initVolumeMonitor();

    puzzle.dlxLoad();

    initOrderingHeuristics();

    // puzzle.dumpShapes(std::cout);
    if(info)
        dumpInfo();
    enableSigHandler();
}


void Solver::cleanup()
{
    if(solveCleanupMeter == NULL)
        solveCleanupMeter = PolyPerf::getInstance()->getMeter("solveCleanup", "solve");
    PerformanceMeasurement measurement(solveCleanupMeter);

    disableSigHandler();
    if(info)
        dumpStats();
    deleteStats();
    deleteRemainingShapeList();
    numMobilePieces = 0;
    puzzle.setPuzzle(NULL);
}



void Solver::enableSigHandler()
{
#ifndef NOSIGNALS
    struct sigaction act;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;

    act.sa_handler = setStatsRequest;
    sigaction(10, &act, 0);

    act.sa_handler = setTraceRequest;
    sigaction(12, &act, 0);
#endif
}


void Solver::processSigRequest()
{
#ifndef NOSIGNALS
    sigRequest = false;
    if(statsRequest)
    {
        dumpStats();
        std::cout << std::endl << *PolyPerf::getInstance();
        statsRequest = false;
    }
    if(traceRequest)
    {
        showTrace();
        traceRequest = false;
    }
#endif
}


void Solver::disableSigHandler()
{
#ifndef NOSIGNALS
    struct sigaction act;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;

    act.sa_handler = SIG_DFL;
    sigaction(10, &act, 0);

    act.sa_handler = SIG_DFL;
    sigaction(12, &act, 0);

    statsRequest = false;
    traceRequest = false;
    sigRequest   = false;
#endif
}


void Solver::allocRemainingShapeList()
{
    remainingShapeList  = makeArray<int>(puzzle.getNumMobileShapes());
    remainingShapeCount = makeArray<int>(puzzle.getNumShapes());
}


void Solver::deleteRemainingShapeList()
{
    deleteArray(remainingShapeList, 0);
    deleteArray(remainingShapeCount, 0);
}


void Solver::initControlVariables()
{
    volumeBacktrackControl = volumeBacktrack > 0 ? volumeBacktrack : numMobilePieces + 1;
    parityFilterControl    = translateFilter(parityFilter);
    volumeFilterControl    = translateFilter(volumeFilter);
    fitFilterControl       = translateFilter(fitFilter);
    parityPlaceControl     = parityBacktrack ? 0 : parityFilterControl;
    volumePlaceControl     = std::min(volumeBacktrackControl, volumeFilterControl);
    filterControl          = std::min(std::min(parityFilterControl, volumeFilterControl), fitFilterControl);

    // Only filter soltuions if the user requested it AND there is the
    // possibility of rotationally redundant soltuions in the search space.
    //
    uniqueControl = unique && puzzle.getRedundancy();
}


void Solver::initOrderingHeuristics()
{
    heuristicStore.setDefaultXc         ((puzzle.getXDim() - 1) / 2.0);
    heuristicStore.setDefaultYc         ((puzzle.getYDim() - 1) / 2.0);
    heuristicStore.setDefaultZc         ((puzzle.getZDim() - 1) / 2.0);
    heuristicStore.setNof               (nof);
    heuristicStore.setFila              (fila);
    heuristicStore.setDlxRoot           (puzzle.getDlxRoot());
    heuristicStore.setNumMobileShapes   (puzzle.getNumMobileShapes());
    heuristicStore.setNumMobilePieces   (puzzle.getNumMobilePieces());

    // The following information (the shapeList, the filaGrid, and the
    // occupancy) isn't even defined yet, but it's all pass by const
    // reference, so the heuristics will always have direct access to the most
    // recent data when needed.
    // 
    heuristicStore.setRemainingShapeList(remainingShapeList);
    heuristicStore.setFilaGrid          (puzzle.getFilaGrid());

    std::map<int,std::string> defList = orderingHeuristicConfig.getDefList();

    // If the defList has no entries, or if there is no heuristic defined for
    // the first piece to be placed (i.e., if the last entry in defList has an
    // activation level that is less than the number of mobile pieces), then
    // add an entry to the end of the list to setup the default size ("s")
    // heuristic.
    //
    if(defList.size() == 0 || (*defList.rbegin()).first < numMobilePieces)
        defList[numMobilePieces] = "s";
    std::map<int,std::string>::const_iterator i = defList.begin();
    for(int n = 1; n <= numMobilePieces; ++n)
    {
        while((*i).first < n)
            ++i;
        heuristicStore.setHeuristic(n, (*i).second);
    }
}

int Solver::translateFilter(int filterSetting)
{
    switch(filterSetting)
    {
        case FILTER_OFF:
            return numMobilePieces + 1;
        case FILTER_ONCE:
            return numMobilePieces;
        default:
            return filterSetting;
    }
};


std::string Solver::formatFilter(int filterLevel)
{
    switch(filterLevel)
    {
        case FILTER_OFF:
            return "OFF";
        case FILTER_ONCE:
            return "ONCE";
        default:
            return lexical_cast<std::string>(filterLevel);
    }
}


void Solver::initStats()
{
    solutionCount        = 0;
    redundantCount       = 0;
    traceCount           = 0;
#ifndef NO_POLY_STATS
    parityBacktrackCount = initArray<long long>(numMobilePieces+1, 0);
    fitFilterCount       = initArray<long long>(numMobilePieces+1, 0);
    parityFilterCount    = initArray<long long>(numMobilePieces+1, 0);
    volumeBacktrackCount = initArray<long long>(numMobilePieces+1, 0);
    volumeFilterCount    = initArray<long long>(numMobilePieces+1, 0);
    fitsList             = initArray<long long>(numMobilePieces,   0);
    attemptsList         = initArray<long long>(numMobilePieces,   0);
#endif
}


void Solver::deleteStats()
{
#ifndef NO_POLY_STATS
    deleteArray(parityBacktrackCount, 0);
    deleteArray(fitFilterCount, 0);
    deleteArray(parityFilterCount, 0);
    deleteArray(volumeBacktrackCount, 0);
    deleteArray(volumeFilterCount, 0);
    deleteArray(fitsList, 0);
    deleteArray(attemptsList, 0);
#endif
}


int Solver::redundancyFilterNameToIndex()
{
    if(redundancyFilterName == REDUNDANCY_FILTER_OFF_NAME)
        return REDUNDANCY_FILTER_OFF_INDEX;

    if(redundancyFilterName == REDUNDANCY_FILTER_AUTO_NAME)
        return REDUNDANCY_FILTER_AUTO_INDEX;

    for(int si = 0; si < puzzle.getNumShapes(); ++si)
    {
        const Shape& s = *puzzle.getShapes()[si];
        for(int ii = 0; ii < s.getNumCopies(); ++ii)
            if(s.getPiece(ii)->getName() == redundancyFilterName)
                return si;
    }

    std::ostringstream errMsg;
    errMsg << "***Redundancy Filter Processing Error:\n\n"
        "Specified filter piece " << lexical_cast<std::string>(redundancyFilterName) << " not found.";
    throw std::runtime_error(errMsg.str());
}


void Solver::dumpInfo()
{
    std::cout << std::endl << "# Puzzle size:" << std::endl <<
        "DIMENSION=" << puzzle.getXDim() << "x" << puzzle.getYDim() <<
        "x" << puzzle.getZDim() << std::endl;

    std::cout << std::endl << "# Shape list:" << std::endl;
    for(int si = 0; si < puzzle.getNumShapes(); ++si)
    {
        std::cout << "SHAPE[" << si << "]={";
        std::cout << *puzzle.getShapes()[si] << "}" << std::endl;
    }

    std::cout << std::endl <<
        "# One sided polyomino constraint:" << std::endl <<
        "ONE_SIDE=" << (puzzle.getOneSide() ? "ON" : "OFF") << std::endl;

    std::cout << std::endl <<
        "# Redundancy filter piece name:" << std::endl <<
        "REDUNDANCY_FILTER=" << (redundancyFilterIndex == REDUNDANCY_FILTER_OFF_INDEX ? std::string("OFF") : puzzle.getShapes()[redundancyFilterIndex]->getPiece()->getName()) << std::endl;

    std::cout << std::endl <<
        "# Piece rotations and translations that fit in the solution space:" << std::endl <<
        "BOUNDED=" << puzzle.getBoundedImageCount() << std::endl;

    // Bug in unbounded count when redundancy filter is enabled.
    // See RELEASE_NOTES.txt
    //
    if(redundancyFilterName == REDUNDANCY_FILTER_OFF_NAME)
    {
        std::cout << std::endl <<
            "# Piece rotations and translations that don't fit in the solution space:" << std::endl <<
            "UNBOUNDED=" << puzzle.getUnboundedImageCount() << std::endl;
    }

    std::cout << std::endl <<
        "# Are solutions guaranteed to be rotationally unique without filtering?" << std::endl <<
        "ROTATIONALLY_UNIQUE=" << (puzzle.getRedundancy() ? "false" : "true") << std::endl;

    std::cout << std::endl <<
        "# Parity of solution space (after all fixed-position pieces are placed):" << std::endl <<
        "PUZZLE_PARITY=" << puzzle.getUntiledParity() << std::endl;
}


void Solver::solve()
{
    if(solveAlgoMeter == NULL)
    {
        solveAlgoMeter    = PolyPerf::getInstance()->getMeter("solveAlgo",    "solve");
        solveFilterMeter  = PolyPerf::getInstance()->getMeter("solveFilter",  "solveAlgo");
        filaMeter         = PolyPerf::getInstance()->getMeter("fila",         "solveAlgo");
        filaInitMeter     = PolyPerf::getInstance()->getMeter("filaInit",     "fila");
        filaAlgoMeter     = PolyPerf::getInstance()->getMeter("filaAlgo",     "fila");
        filaCleanupMeter  = PolyPerf::getInstance()->getMeter("filaCleanup",  "fila");
    }
    PerformanceMeasurement measure(solveAlgoMeter);

    if(parityBacktrack && !puzzle.checkParity())
    {
#ifndef NO_POLY_STATS
        ++parityBacktrackCount[numMobilePieces];
#endif
        return;
    }

    if(numMobilePieces >= volumeBacktrackControl && !puzzle.checkVolume())
    {
#ifndef NO_POLY_STATS
        ++volumeBacktrackCount[numMobilePieces];
#endif
        return;
    }

    if(monteCarlo.getNumTrials() > 0)
    {
        Random random(monteCarlo.getSeed());
        for(monteCarloTrial = 0; monteCarloTrial < monteCarlo.getNumTrials(); ++monteCarloTrial)
        {
            puzzle.dlxRandomize(random);
            solveOnce();
            endTrialControl = false;
        }
    }
    else
    {
        solveOnce();
    }
}


void Solver::solveOnce()
{
    int nrp = puzzle.getNumRemainingPieces();
    int numFilter = 0;
    if(nrp >= filterControl)
        numFilter = dlxFilter(nrp, 1);

    if(puzzle.getNumRemainingPieces() <= fila && puzzle.getNumRemainingGridPoints() <= maxFilaGridPoints)
        solveFila(nrp);
    else
        solveDlx_outter(heuristicStore.selectDlx(nrp), numMobilePieces);

    if(nrp >= filterControl)
        dlxUnfilter(numFilter);
}


int Solver::dlxFilter(int nrp, int parityOfLastPiecePlaced)
{
    PerformanceMeasurement measure(solveFilterMeter);

    int numFilter = 0;

    // Do volume and parity pruning first as these can filter images required
    // for the fit checks below to pass (resulting in more fit pruning).  This
    // is probably rare, but I can definitely construct cases where this
    // happens.

    // It's pointless to look for rows that can be filterd due to parity
    // constraint violations if the new parity is equal to a parity that
    // has already been filtered.
    //
    if(nrp >= parityFilterControl && parityOfLastPiecePlaced)
    {
        int np = puzzle.dlxFilterParity();
        numFilter += np;
#ifndef NO_POLY_STATS
        parityFilterCount[nrp] += np;
#endif
    }

    if(nrp >= volumeFilterControl)
    {
        int np = puzzle.dlxFilterVolume();
        numFilter += np;
#ifndef NO_POLY_STATS
        volumeFilterCount[nrp] += np;
#endif
    }

    if(nrp >= fitFilterControl)
    {
        int np = puzzle.dlxFilterFit();
        numFilter += np;
#ifndef NO_POLY_STATS
        fitFilterCount[nrp] += np;
#endif
    }

    return numFilter;
}


void Solver::dlxUnfilter(int numUnfilter)
{
    for(int i = 0; i < numUnfilter; ++i)
        puzzle.dlxUnfilterRow();
}


void Solver::solveDlx_outter(DlxHead* target, int nrp)
{
    if(target->numRow == 0)
        return;

#ifndef NOSIGNALS
    if(sigRequest)
        processSigRequest();
#endif

    --nrp;

    // Only cover the target column if the number-of-copies count for that
    // column hits zero.
    //
    if(--target->count == 0)
        puzzle.dlxCover(target);

    for(DlxNode* r = target->down; r != target && !endTrialControl; r = r->down)
    {
        puzzle.placeStack(r->image);

#ifndef NO_POLY_STATS
        ++attemptsList[nrp];
        ++fitsList[nrp];
#endif

        if(trace && (trace > 0 ? nrp >= trace-1 : nrp == -trace-1))
            showTrace();

        if(nrp == goal)
            processSolution();
        else
            solveDlx_parityBacktrack(r, nrp);

        puzzle.unplaceStack();

        if(trace > 0 && nrp >= trace-1)
            showTrace();

        if(monteCarlo.getRange() == nrp + 1)
            endTrialControl = true;
    }

    if(target->count++ == 0)
        puzzle.dlxUncover(target);
}

int Solver::initFila(int nrp)
{
    int nrs = puzzle.initFila(remainingShapeList, remainingShapeCount);
    heuristicStore.loadImages(nrp, puzzle.getNumFilaGridPoints());
    return nrs;
}

void Solver::clearFila()
{
    heuristicStore.clearImages();
    puzzle.clearFila();
}

void Solver::solveFila_outter(int nrp, int nrs, gridbitfield_t occupancy)
{
    //std::cout << "nrp=" << nrp << ", nrs=" << nrs << ", occupancy=" << Bits<gridbitfield_t>(occupancy) << std::endl;
    ImageListSet imageListSet = heuristicStore.selectFila(nrp, nrs, occupancy, false);

    // The S and E heuristics can actually detect no-fit situations and
    // communicate that result by returning a NULL ImageListSet.
    //
    if(imageListSet == NULL)
        return;

#ifndef NOSIGNALS
    if(sigRequest)
        processSigRequest();
#endif

    --nrp;

    /*
    std::cout << "occupancy=" << Bits<gridbitfield_t>(occupancy) << std::endl;
    std::cout << "possibilities:" << std::endl;
    for(int si = 0; si < nrs; ++si)
    {
        int pi = remainingShapeList[si];
        ImageList& imageList = imageListSet[pi];
        ImageList::iterator end = imageList.end();
        for(ImageList::iterator ii = imageList.begin(); ii != end; ++ii)
        {
            Image* i = *ii;
            std::cout << *i;
            if(occupancy & i->layoutMask)
                std::cout << "*";
            std::cout << std::endl;
        }
    }
    */

    for(int si = 0; si < nrs; ++si)
    {
        int s = remainingShapeList[si];
        ImageList& imageList = imageListSet[s];
        if(imageList.size() == 0) // This quick-out improves performance maybe 1-2%.
            continue;

        if(!--remainingShapeCount[s])
            std::swap(remainingShapeList[si], remainingShapeList[--nrs]);

#ifndef NO_POLY_STATS
        attemptsList[nrp] += imageList.size();
#endif

        for(Image* i : imageList)
        {
            if(occupancy & i->layoutMask)
                continue;

            puzzle.placeStack(i);

#ifndef NO_POLY_STATS
            ++fitsList[nrp];
#endif

            if(trace && (trace > 0 ? nrp >= trace-1 : nrp == -trace-1))
                showTrace();

            if(nrp == goal)
                processSolution();
            else
                solveFila_parityBacktrack(nrp, nrs, occupancy, i);

            puzzle.unplaceStack();

            if(trace > 0 && nrp >= trace-1)
                showTrace();
        }

        if(!remainingShapeCount[s]++)
            std::swap(remainingShapeList[si], remainingShapeList[nrs++]);
    }
}


void Solver::processSolution()
{
    ++solutionCount;
    if(!uniqueControl || filterSolution())
    {
        if(!quiet)
        {
            std::cout << std::endl << "# --- SOLUTION " << solutionCount - redundantCount << " ---" << std::endl;
            puzzle.printState(std::cout, outputFormat);
        }
    }
}


bool Solver::filterSolution()
{
    std::vector<PIECEID_T> solution;
    puzzle.getNormalizedState(solution);
    if(allSolutionList.find(solution) != allSolutionList.end())
    {
        ++redundantCount;
        return false;
    }

    std::vector<PIECEID_T> rotatedSolution;
    const std::vector<std::vector<int> >& symmetricPermutationList = puzzle.getSymmetricPermutationList();
    for(std::vector<std::vector<int> >::const_iterator ri = symmetricPermutationList.begin(); ri != symmetricPermutationList.end(); ++ri)
    {
        puzzle.rotateAndNormalizeState(solution, *ri, rotatedSolution);
        allSolutionList.insert(rotatedSolution);
    }

    return true;
}


void Solver::showTrace()
{
    std::cout << "TRACE " << ++traceCount << ":" << std::endl;
    puzzle.printState(std::cout, outputFormat);
    std::cout << "END TRACE" << std::endl;
}


void Solver::dumpStats()
{
    double secs = (double) PolyPerf::getInstance()->getMeter("all")->getTime() / 1e6;

    std::cout <<
        std::endl << "# Number of solutions found:" << std::endl <<
        "SOLUTIONS=" << solutionCount << std::endl;

    std::cout <<
        std::endl << "# Number of solutions found per second:" << std::endl <<
        "SOLUTIONS_PER_SEC=";
    if(secs > 0)
        std::cout << std::setprecision(9) << solutionCount / secs;
    else
        std::cout << "*";
    std::cout << std::endl;

    if(!puzzle.getRedundancy() || unique)
    {
        std::cout <<
            std::endl << "# Number of rotationally unique solutions found:" << std::endl <<
            "UNIQUE=" << (solutionCount - redundantCount) << std::endl;

        std::cout <<
            std::endl << "# Number of rotationally unique solutions found per second:" << std::endl <<
            "UNIQUE_PER_SEC=";
        if(secs > 0)
            std::cout << std::setprecision(9) << (solutionCount - redundantCount) / secs;
        else
            std::cout << "*";
        std::cout << std::endl;
        
        std::cout <<
            std::endl << "# Number of rotationally redundant solutions found but filtered out:" << std::endl <<
            "REDUNDANT=" << redundantCount << std::endl;
    }
    else
    {
        std::cout <<
            std::endl << "# Number of rotationally unique solutions found:" << std::endl <<
            "UNIQUE=" << "UNKNOWN" << std::endl;
    }

    if(monteCarlo.getNumTrials() > 0)
    {
        std::cout <<
            std::endl << "# Number of Monte-Carlo trials completed:" << std::endl <<
            "MONTE_CARLO_TRIAL=" << monteCarloTrial << std::endl;
    }

#ifndef NO_POLY_STATS

    long long sum = 0;
    for(int ci = 0; ci <= numMobilePieces; ++ci)
        sum += parityBacktrackCount[ci];

    std::cout <<
        std::endl << "# Number of backtracks due to parity constraint:" << std::endl <<
        "PARITY_BACKTRACK_TOTAL=" << sum << std::endl;

    sum = 0;
    for(int ci = 0; ci <= numMobilePieces; ++ci)
        sum += volumeBacktrackCount[ci];

    std::cout <<
        std::endl << "# Number of backtracks due to volume constraint:" << std::endl <<
        "VOLUME_BACKTRACK_TOTAL=" << sum << std::endl;

    sum = 0;
    for(int ci = 0; ci <= numMobilePieces; ++ci)
        sum += fitFilterCount[ci];

    std::cout <<
        std::endl << "# Number of rows filtered due to fit constraint:" << std::endl <<
        "FIT_FILTER_TOTAL=" << sum << std::endl;

    sum = 0;
    for(int ci = 0; ci <= numMobilePieces; ++ci)
        sum += parityFilterCount[ci];

    std::cout <<
        std::endl << "# Number of rows filtered due to parity constraint:" << std::endl <<
        "PARITY_FILTER_TOTAL=" << sum << std::endl;

    sum = 0;
    for(int ci = 0; ci <= numMobilePieces; ++ci)
        sum += volumeFilterCount[ci];

    std::cout <<
        std::endl << "# Number of rows filtered due to volume constraint:" << std::endl <<
        "VOLUME_FILTER_TOTAL=" << sum << std::endl;

    sum = 0;
    for(int ci = 0; ci < numMobilePieces; ++ci)
        sum += attemptsList[ci];

    std::cout << std::endl << "# Number of times pieces were attempted to be placed:" << std::endl <<
        "ATTEMPTS_TOTAL=" << sum << std::endl;

    sum = 0;
    for(int ci = 0; ci < numMobilePieces; ++ci)
        sum += fitsList[ci];

    std::cout << std::endl << "# Number of times pieces fit:" << std::endl <<
        "FITS_TOTAL=" << sum << std::endl;

    std::cout << std::endl << "# Number of DLX list deletions when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "DLX_LIST_DELETES[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << (puzzle.getDanceCount(ci)) << std::endl;

    std::cout << std::endl << "# Number of times a piece was placed with a parity moving away from the target parity when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= puzzle.getNumPieces(); ++ci)
        std::cout << "WRONG_WAY_PARITY_WALK[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << (puzzle.getWrongWayParityCount(ci)) << std::endl;

    std::cout << std::endl << "# Number of backtracks due to parity constraint when N pieces were left to be placed:" << std::endl;

    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "PARITY_BACKTRACK[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << parityBacktrackCount[ci] << std::endl;

    std::cout << std::endl << "# Number of backtracks due to volume constraint when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "VOLUME_BACKTRACK[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << volumeBacktrackCount[ci] << std::endl;

    std::cout << std::endl << "# Number of filters due to parity constraint when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "PARITY_FILTER[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << parityFilterCount[ci] << std::endl;

    std::cout << std::endl << "# Number of filters due to fit constraint when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "FIT_FILTER[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << fitFilterCount[ci] << std::endl;

    std::cout << std::endl << "# Number of filters due to volume constraint when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "VOLUME_FILTER[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << volumeFilterCount[ci] << std::endl;

    std::cout << std::endl << "# Number of placement attempts when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "ATTEMPTS[" << std::setw(2) << ci << "]=" <<
            std::setw(14) << attemptsList[ci-1] << std::endl;

    std::cout << std::endl << "# Number of fits when N pieces were left to be placed:" << std::endl;
    for(int ci = 1; ci <= numMobilePieces; ++ci)
        std::cout << "FITS[" << std::setw(2)  << ci << "]=    " <<
            std::setw(14) << fitsList[ci-1] << std::endl;

#endif
}
