// polycube:  A polyomino and polycube puzzle solver.
// Copyright 2011-2018 Matthew T. Busche.
//
// This program is free software:  you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along
// with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef SOLVER_HPP
#define SOLVER_HPP

#include <algorithm>
#include <iostream>
#include <iomanip>
#include "time.h"
#include "fastbits.hpp"
#include "MonteCarloConfig.hpp"
#include "OrderingHeuristicConfig.hpp"
#include "OrderingHeuristicStore.hpp"
#include "PuzzleSolverInterface.hpp"

// Variable Naming Notes:
//
//    nrp = num remaining pieces
//    nrs = num remaining shapes
//

class Solver
{
    private:

        // ####################################################
        // CONFIG SETTINGS
        // ####################################################

        /** The solver is setup to slower but smarter techniques when there
         ** are still many pieces to place in the puzzle solution space, but
         ** to use a simpler tiling technique when there are fewer pieces
         ** remaining to be placed.  The last fila pieces are placed using a
         ** simpler FILA algorithm.
         **/
        int fila;

        /** Specifies the format of solution and trace output.
         **/ 
        OutputFormatConfig outputFormat;

        /** The minimum number of remaining pieces for fit filters.
         **/
        int fitFilter;

        /** Each time the number of remaining pieces equals the goal, a
         ** solution is declared and processed, and a backtrack is forced
         ** (even if additional pieces could be placed).
         **/
        int goal;

        /** If true, various informational messages are printed.  These
         ** include puzzle input information; and solution, and performance
         ** statistics.
         **/
        bool info;

        MonteCarloConfig monteCarlo;

        OrderingHeuristicConfig orderingHeuristicConfig;

        bool nof;

        /** If true, a parity constraint check is made after each piece
         ** placement.
         **/
        bool parityBacktrack;

        /** The minimum number of remaining pieces for parity filters.
         **/
        int parityFilter;

        /** If true, solutions are not output.
         **/
        bool quiet;

        /** The name of a piece in pieceList which the solver constrains
         ** rotations and/or translations to eliminate rotationally equivalent
         ** solutions during the search.  The value REDUNDANCY_FILTER_AUTO_NAME
         ** causes the solver to make a good choice for which piece to limit
         ** rotations.  The value REDUNDANCY_FILTER_OFF_NAME turns off
         ** constraint processing.
         **/
        std::string redundancyFilterName;

        /** The index of the piece whose images are filtered to eliminate
         ** rotational redundancy among solutions.  The value
         ** REDUNDANCY_FILTER_OFF_INDEX turns this feature off.  This field is
         ** not directly settable by the user, but is rather derived from the
         ** user settable field constrainName during initialization.
         **/
        int redundancyFilterIndex;

        /** If set to true, the DLX column corresponding to the piece selected
         ** for rotational redundancy filtering is forcibly selected to be
         ** processed first.
         **/
        bool redundancyFilterFirst;

        /** If trace is set to some number K, then after each piece placement,
         ** if there are at least K pieces remaining to be placed, then the
         ** state of the puzzle is displayed.  Also, after each piece removal
         ** if there are at least K+1 pieces remaining to be placed, then the
         ** state of the puzzle is displayed.  If K is negative, then the state
         ** of the puzzle is displayed only after a piece is placed and there
         ** are exactly -K pieces left to place.  A zero value turns this
         ** feature off.
         **/
        int trace;

        /** If true, rotated copies of previously discovered rotationally
         ** unique solutions are maintained, and only rotationlly unique
         ** solutions are output.
         **/
        bool unique;

        /** The minimum number of remaining pieces for volume backtracks.
         **/
        int volumeBacktrack;

        /** The minimum number of remaining pieces for volume filters.
         **/
        int volumeFilter;

        // ####################################################################
        // CONTROL VARIABLES.
        //
        // These variables are all just simple translations of configuration
        // variables or combinations of configuration variables that simplify
        // for the algorithms the decision making process of whether or not to
        // invoke some image filter, backtrack processing, trace, etc.
        // ####################################################################

        /** Same as volumeBacktrack but a 0 value which means OFF is
         ** tranlated to a number greater than the number of pieces
         ** in the puzzle, simplifying the conditional used
         ** to determine if a volume check should be made.
         **/
        int volumeBacktrackControl;

        /** Same as fitFilter but special constants FILTER_OFF and FILTER_ONCE
         ** are translated to actual piece counts for easier processing.  This
         ** field is not directly settable by the user, but is rather derived
         ** from the user settable field fitFilter during initialization.
         **/
        int fitFilterControl;

        /** Same as parityFilter but special constants FILTER_OFF and
         ** FILTER_ONCE are translated to actual piece counts for easier
         ** processing.  This field is not directly settable by the user, but
         ** is rather derived from the user settable field parityFilter during
         ** initialization.
         **/
        int parityFilterControl;

        /** Same as volumeFilter but special constants FILTER_OFF and
         ** FILTER_ONCE are translated to actual piece counts for easier
         ** processing.  This field is not directly settable by the user, but
         ** is rather derived from the user settable field volumeFilter during
         ** initialization.
         **/
        int volumeFilterControl;

        /** This is just a convenience lookup to determine if the image set
         ** should be filtered after image placement:
         **
         **   filterControl = min(fitFilterControl, volumeFilterControl, parityFilterControl)
         **/
        int filterControl;

        /** The minimum number of remaining pieces needed to continue
         ** maintaining the parity state.  Once the number of remaining pieces
         ** falls below this threshold, neither parity filters nor parity
         ** backtracks are used and it becomes pointless to continue to
         ** maintain the parity state.  Parity state updates are so simple we
         ** could conceivably be better off by just always updating the parity
         ** state; but I've kept this control for the aesthetics of symmetry
         ** with volumePlaceControl.
         **/
        int parityPlaceControl;

        /** The minimum of volumeFilterControl and volumeBacktrackControl.
         ** This variable is used to determine whether to continue maintaining
         ** the GridPoint fill state as pieces are placed.  Once the number of
         ** remaining pieces falls below this threshold, neither volume filters
         ** nor volume backtracks are used and it becomes pointless to
         ** continue to maintain the fill state.
         **/
        int volumePlaceControl;

        /** Nominally this will have the same value as unique; however, if
         ** either the Puzzle has no symmetric rotations, or if the Puzzle has
         ** determined that the redundancyFilter eliminates all rotationally
         ** redundant solutions from the search space; then uniqueControl will
         ** necessarily be false.
         **/
        int uniqueControl;

        /** Flag is set when a Monte Carlo trial is ended, forcing DLX to
         ** unwind.
         **/
        bool endTrialControl;

        // ####################################################
        // OTHER STUFF
        // ####################################################

        /** If the soltuion filter is enabled, then each time a rotationally
         ** unique solution is found, it is rotated by all rotations in the
         ** Puzzle's symmetric rotation list.  Each of these rotated solutions
         ** are then added to allSolutionList.  Each time a solution is found,
         ** it is checked against allSolutionList to determine if it is
         ** rotationally unique from previously discovered solutions.  This
         ** list and all associated processing is forcibly disabled if the
         ** Puzzle determines that no rotationally redundant solutions will be
         ** found.
         **/
        std::set<std::vector<PIECEID_T>, StateComp> allSolutionList;

        OrderingHeuristicStore heuristicStore;

        /** This is just used as an iteration variable for monte-carlo
         ** anlaysis, but I made it a class member so I could print the
         ** current trial number as part of the status dump (which can be
         ** triggered asynchronously via unix signal 10).
         **/
        int monteCarloTrial;

        // ####################################################
        // THE PUZZLE AND SOME OF IT'S IMPORTANT PROPERTIES
        // ####################################################

        /** An interface wrapper around the Puzzle object upon which this
         ** solver is operating.
         **/
        PuzzleSolverInterface puzzle;

        /** The number of individual mobile pieces in the puzzle.  (This is
         ** the same as numMobileShapes if all mobile pieces have a unique
         ** shape.)  Copied here from puzzle for convenience.
         **/
        int numMobilePieces;

        /** The array of indicees of mobile pieces that are still free.  This
         ** is reset each time DLX is turned off and FILA is turned on.
         **/
        int* remainingShapeList;

        /** remainingShapeCount[p] gives the number of copies of the shape
         ** with id p that is still free.  This is reset each time Dlx is
         ** turned off and FILA is turned on.
         **/
        int* remainingShapeCount;

        // ####################################################
        // STATISTICS
        // ####################################################

        unsigned long long traceCount;
        long long  solutionCount;
        long long  redundantCount;
#ifndef NO_POLY_STATS
        long long* parityBacktrackCount;
        long long* parityFilterCount;
        long long* volumeBacktrackCount;
        long long* volumeFilterCount;
        long long* fitFilterCount;
        long long* attemptsList;
        long long* fitsList;
#endif

        PerformanceMeter* solveMeter;
        PerformanceMeter* solveInitMeter;
        PerformanceMeter* solveAlgoMeter;
        PerformanceMeter* solveFilterMeter;
        PerformanceMeter* filaMeter;
        PerformanceMeter* filaInitMeter;
        PerformanceMeter* filaAlgoMeter;
        PerformanceMeter* filaCleanupMeter;
        PerformanceMeter* solveCleanupMeter;

    public:

        Solver();
        int getFila() const;
        void setFila(int fila);
        const OutputFormatConfig& getOutputFormat() const;
        void setOutputFormat(const OutputFormatConfig& outputFormat);
        int getFitFilter() const;
        void setFitFilter(int fitFilter);
        int getGoal() const;
        void setGoal(int goal);
        bool getInfo() const;
        void setInfo(bool info);
        bool getNof() const;
        void setNof(bool nof);
        const MonteCarloConfig& getMonteCarlo() const;
        void setMonteCarlo(const MonteCarloConfig& monteCarlo);
        const OrderingHeuristicConfig& getOrderingHeuristicConfig() const;
        void setOrderingHeuristicConfig(const OrderingHeuristicConfig& orderingHeuristicConfig);
        bool getParityBacktrack() const;
        void setParityBacktrack(bool parityBacktrack);
        int getParityFilter() const;
        void setParityFilter(int parityFilter);
        int getQuiet() const;
        void setQuiet(bool quiet);
        const std::string& getRedundancyFilterName() const;
        void setRedundancyFilterName(const std::string& redundancyFilterName);
        bool getRedundancyFilterFirst() const;
        void setRedundancyFilterFirst(bool redundancyFilterFirst);
        int getTrace();
        void setTrace(int trace);
        bool getUnique() const;
        void setUnique(bool unique);
        int getVolumeBacktrack() const;
        void setVolumeBacktrack(int volumeBacktrack);
        int getVolumeFilter() const;
        void setVolumeFilter(int volumeFilter);
        void solve(Puzzle& puzzle);

    private:

        void init();
        void cleanup();
        int translateFilter(int filterLevel);
        static std::string formatFilter(int filterSetting);
        void enableSigHandler();
        void disableSigHandler();
        void processSigRequest();
        void allocRemainingShapeList();
        void deleteRemainingShapeList();
        void createPerformanceMeters();
        void initControlVariables();
        void initOrderingHeuristics();
        void deleteOrderingHeuristics();
        void initStats();
        void deleteStats();
        int redundancyFilterNameToIndex();
        void dumpInfo();
        void solve();
        void solveOnce();
        int dlxFilter(int nrp, int parityOfLastPiecePlaced);
        void dlxUnfilter(int numUnfilter);

        // These next five functions together form the basic top-level DLX
        // search algorithm.  They call eachother in a recursive chain:
        //
        //   solveDlx_outter calls
        //   solveDlx_parityBacktrack which calls
        //   solveDlx_volumeBacktrack which calls
        //   solveDlx_coverAndFilter which calls
        //   solveDlx_inner which calls
        //   solveDlx_outter
        //
        // It could easily all be written as single recursive function but
        // this approach has less branching and is perhaps easier to follow.

        void solveDlx_outter(DlxHead* target, int nrp);
        void solveDlx_parityBacktrack(DlxNode* r, int nrp);
        void solveDlx_volumeBacktrack(DlxNode* r, int nrp);
        void solveDlx_coverAndFilter(DlxNode* r, int nrp);
        void solveDlx_inner(DlxNode* r, int nrp);

        int initFila(int nrp);
        void clearFila();

        void solveFila                (int nrp);
        void solveFila_outter         (int nrp, int nrs, const gridbitfield_t occupancy);
        void solveFila_parityBacktrack(int nrp, int nrs, const gridbitfield_t occupancy, Image* i);
        void solveFila_volumeBacktrack(int nrp, int nrs, const gridbitfield_t occupancy, Image* i);

        void processSolution();

        /** Checks to see if solution is unique from those in allSolutionList.
         ** If not then the redundant solution count is incremented and false
         ** is returned.  Otherwise, that solution is added to
         ** uniqueSolutionList, all rotations of that solution (that result in
         ** the same puzzle shape) are added to allSolutionList, and true is
         ** returned.
         **/
        bool filterSolution();
        void showTrace();
        void dumpStats();
};


inline int Solver::getFila() const
{
    return fila;
}

inline void Solver::setFila(int fila)
{
    this->fila = fila;
}

inline const OutputFormatConfig& Solver::getOutputFormat() const
{
    return outputFormat;
}

inline void Solver::setOutputFormat(const OutputFormatConfig& outputFormat)
{
    this->outputFormat = outputFormat;
}

inline int Solver::getFitFilter() const
{
    return fitFilter;
}

inline void Solver::setFitFilter(int fitFilter)
{
    this->fitFilter = fitFilter;
}

inline int Solver::getGoal() const
{
    return goal;
}

inline void Solver::setGoal(int goal)
{
    this->goal = goal;
}

inline bool Solver::getInfo() const
{
    return info;
}

inline void Solver::setInfo(bool info)
{
    this->info = info;
}

inline bool Solver::getNof() const
{
    return nof;
}

inline void Solver::setNof(bool nof)
{
    this->nof = nof;
}

inline const MonteCarloConfig& Solver::getMonteCarlo() const
{
    return monteCarlo;
}

inline void Solver::setMonteCarlo(const MonteCarloConfig& monteCarlo)
{
    this->monteCarlo = monteCarlo;
}

inline const OrderingHeuristicConfig& Solver::getOrderingHeuristicConfig() const
{
    return orderingHeuristicConfig;
}

inline void Solver::setOrderingHeuristicConfig(const OrderingHeuristicConfig& orderingHeuristicConfig)
{
    this->orderingHeuristicConfig = orderingHeuristicConfig;
}

inline bool Solver::getParityBacktrack() const
{
    return parityBacktrack;
}

inline void Solver::setParityBacktrack(bool parityBacktrack)
{
    this->parityBacktrack = parityBacktrack;
}

inline int Solver::getParityFilter() const
{
    return parityFilter;
}

inline void Solver::setParityFilter(int parityFilter)
{
    this->parityFilter = parityFilter;
}

inline int Solver::getQuiet() const
{
    return quiet;
}

inline void Solver::setQuiet(bool quiet)
{
    this->quiet = quiet;
}

inline const std::string& Solver::getRedundancyFilterName() const
{
    return redundancyFilterName;
}

inline void Solver::setRedundancyFilterName(const std::string& redundancyFilterName)
{
    this->redundancyFilterName = redundancyFilterName;
}

inline bool Solver::getRedundancyFilterFirst() const
{
    return redundancyFilterFirst;
}

inline void Solver::setRedundancyFilterFirst(bool redundancyFilterFirst)
{
    this->redundancyFilterFirst = redundancyFilterFirst;
}

inline int Solver::getTrace()
{
    return trace;
}

inline void Solver::setTrace(int trace)
{
    this->trace = trace;
}

inline bool Solver::getUnique() const
{
    return unique;
}

inline void Solver::setUnique(bool unique)
{
    this->unique = unique;
}

inline int Solver::getVolumeBacktrack() const
{
    return volumeBacktrack;
}

inline void Solver::setVolumeBacktrack(int volumeBacktrack)
{
    this->volumeBacktrack = volumeBacktrack;
}

inline int Solver::getVolumeFilter() const
{
    return volumeFilter;
}

inline void Solver::setVolumeFilter(int volumeFilter)
{
    this->volumeFilter = volumeFilter;
}


inline void Solver::solveDlx_parityBacktrack(DlxNode* r, int nrp)
{
    if(nrp >= parityPlaceControl)
    {
        puzzle.placeParity(r->image->parity);

        // I could modify this to skip the parity violation check if parity
        // filtering was enabled on the previous step (as I do for the volume
        // violation check below), but the check is just an an array lookup,
        // so I think it would actually slow things down on average.
        //
        if(!parityBacktrack || puzzle.checkParity())
            solveDlx_volumeBacktrack(r, nrp);
#ifndef NO_POLY_STATS
        else
            ++parityBacktrackCount[nrp];
#endif
        puzzle.unplaceParity(r->image->parity);
    }
    else
        solveDlx_volumeBacktrack(r, nrp);
}


inline void Solver::solveDlx_volumeBacktrack(DlxNode* r, int nrp)
{
    if(nrp >= volumePlaceControl)
    {
        puzzle.placeVolume(r->image);

        // Here we only continue if the puzzle has no volume constraint
        // violations.  But we skip the volume check (and unconditionally
        // continue) if either:
        //
        //    o the number of remaining pieces is less than the control
        //      setting where volume checks are no longer made, or
        //
        //    o a volume filter was performed at the previous recursion level.
        //      (If ALL images were filtered in the previous step, then it's
        //      guaranteed that the one we're actually using in this step
        //      does not produce a volume constraint violation.)
        //
        if(nrp < volumeBacktrackControl ||
            nrp >= volumeFilterControl-1 ||
            puzzle.checkVolume())
        {
            solveDlx_coverAndFilter(r, nrp);
        }
#ifndef NO_POLY_STATS
        else
            ++volumeBacktrackCount[nrp];
#endif
        puzzle.unplaceVolume(r->image);
    }
    else
        solveDlx_coverAndFilter(r, nrp);
}


inline void Solver::solveDlx_coverAndFilter(DlxNode* r, int nrp)
{
    for(DlxNode* j = r->right; j != r; j = j->right)
        if(--j->head->count == 0)
            puzzle.dlxCover(j->head);

    if(nrp >= filterControl)
    {
        int numFilter = dlxFilter(nrp, r->image->parity);
        solveDlx_inner(r, nrp);
        dlxUnfilter(numFilter);
    }
    else
        solveDlx_inner(r, nrp);

    for(DlxNode* j = r->left; j != r; j = j->left)
        if(j->head->count++ == 0)
            puzzle.dlxUncover(j->head);
}


inline void Solver::solveDlx_inner(DlxNode* r, int nrp)
{
    puzzle.placeGridPointCount(r->image);

    // If the number of remaining pieces has reached the user selected FILA
    // range then we want to switch to FILA; however, there are two situations
    // where I delay this switch:
    //
    //    1) If  the number of remaining holes is larger than the number of
    //       bits in data type gridbitfield_t, then there is no choice but to
    //       continue with DLX.
    //
    //    2) If the branch factor for the target selected by the DLX ordering
    //       heuristic is 0 or 1, then the transition to fila is delayed.
    //
    // The second rule may seem arbitrary, but I don't think it can have a
    // severe negative consequences if included, and I found that for some
    // solve configurations the result of taking it out can be devastating.
    // For example, removing this check results in the following solve to take
    // more than 70% longer:
    //
    //     polycube -n -V -rL -f12 -of=12 -q def/pentominoes_onesided_18x5.txt
    //
    // I think the problem with this solve is that the DLX size heuristic can
    // leave a half-hazard topolgy in it's wake, and the fletcher style
    // left-to-right fill will ignore very hard or impossible to fill regions
    // until it stumbles upon them.
    //
    // Although I think it's good behavior from a performance perspective,
    // it's confusing, difficult to explain, and can make fair comparisons of
    // other algorithm features difficult.  I should probably at least add a
    // way to turn this off!  (Another TODO.)
    //
    DlxHead* target = heuristicStore.selectDlx(nrp);
    if(nrp <= fila && puzzle.getNumRemainingGridPoints() <= maxFilaGridPoints && target->numRow > 1)
        solveFila(nrp);
    else
        solveDlx_outter(target, nrp);

    puzzle.unplaceGridPointCount(r->image);
}


inline void Solver::solveFila(int nrp)
{
    int nrs;
    PerformanceMeasurement measure(filaMeter);
    {
        PerformanceMeasurement measure(filaInitMeter);
        nrs = initFila(nrp);
    }
    {
        PerformanceMeasurement measure(filaAlgoMeter);

        // Because ordering heuristics often take the binary one's complement of
        // the occupancy state to produce a bitfield representing the 'holes'
        // (i.e., unoccupied cells) in the puzzle, it is convenient if the unused
        // bits in occupancy are set to 1.
        //
        solveFila_outter(nrp, nrs, FastBits::upperMask<gridbitfield_t>(puzzle.getNumFilaGridPoints()));
    }
    {
        PerformanceMeasurement measure(filaCleanupMeter);
        clearFila();
    }
}


inline void Solver::solveFila_parityBacktrack(int nrp, int nrs, const gridbitfield_t occupancy, Image* i)
{
    if(nrp >= parityPlaceControl)
    {
        puzzle.placeParity(i->parity);
        if(!parityBacktrack || puzzle.checkParity())
            solveFila_volumeBacktrack(nrp, nrs, occupancy, i);
#ifndef NO_POLY_STATS
        else
            ++parityBacktrackCount[nrp];
#endif
        puzzle.unplaceParity(i->parity);
    }
    else
        solveFila_volumeBacktrack(nrp, nrs, occupancy, i);
}


inline void Solver::solveFila_volumeBacktrack(int nrp, int nrs, const gridbitfield_t occupancy, Image* i)
{
    if(nrp >= volumePlaceControl)
    {
        puzzle.placeVolume(i);
        if(nrp < volumeBacktrackControl || puzzle.checkVolume())
            solveFila_outter(nrp, nrs, occupancy | i->layoutMask);
#ifndef NO_POLY_STATS
        else
            ++volumeBacktrackCount[nrp];
#endif
        puzzle.unplaceVolume(i);
    }
    else
        solveFila_outter(nrp, nrs, occupancy | i->layoutMask);
}


#endif
